Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import Lambda.
Require Import BooleanIndistinguishability.
Require Import Classical.

Set Implicit Arguments.

Definition boolean_sem : SEM atom := @bool_Lequiv.

Lemma good_LASemantics_boolean_sem : good_LASemantics boolean_sem.
Proof.
intros A L LA. splits.
* intros a a' l H atom1 atom2. apply bool_Lequiv_auth_antimonotone.  assumption.
* intros a l l' H atom1 atom2. apply bool_Lequiv_label_antimonotone. assumption.
* intros a l l' H atom1 atom2. apply bool_Lequiv_injection. assumption.
Qed.
Hint Resolve good_LASemantics_boolean_sem.

(** * Characterization of sound LA-maps. *)
Section Soundness_characterization.

Context {A1 L1 A2 L2}
        (LA1: LabelAlgebra A1 L1)
        (LA2: LabelAlgebra A2 L2)
        (m: LA1 ==> LA2).

Definition soundness_prop :=
  (forall a l1 l2, Lmap m l1 ⊑[Amap m a] Lmap m l2 -> l1 ⊑[a] l2)
  /\ (forall l1 l2, l1 =L l2 -> Lmap m l1 =L Lmap m l2).

Lemma soundness_necessary : LAMapSound m boolean_sem -> soundness_prop.
Proof.
intro Hsound. split.
* intros a l1 l2 H. apply NNPP. intro Hnleq.
  assert (bool_Lequiv a l2 (Atom (VBool true) l1) (Atom (VBool false) l1))
    as HLequiv.
  { intros Hlab; tauto. }
  specialize (Hsound _ _ _ _ HLequiv). simpl in Hsound.
  specialize (Hsound (or_introl H)).
  destruct Hsound. congruence.
* intros l1 l2 H.
  assert (bool_Lequiv bottom l2 (Atom (VBool true) l1) (Atom (VBool true) l2))
    as HLequiv.
  { intro Hlab. tauto. }
  specialize (Hsound _ _ _ _ HLequiv). simpl in Hsound.
  destruct Hsound. right; auto. auto.
Qed.

Lemma soundness_sufficient : soundness_prop -> LAMapSound m boolean_sem.
Proof.
intros [HflowsTo Heq] auth lab [v1 l1] [v2 l2] HLequiv.
destruct v1; destruct v2; simpl in HLequiv; try tauto.
intro Hlab.
assert (l1 ⊑[ auth]lab \/ l2 ⊑[ auth]lab) as Hlab'.
{ destruct Hlab; [left | right]; auto. }
specialize (HLequiv Hlab'). destruct HLequiv; auto.
Qed.

Lemma soundness_iff : LAMapSound m boolean_sem <-> soundness_prop.
Proof.
split.
* apply soundness_necessary.
* apply soundness_sufficient.
Qed.

End Soundness_characterization.

(** * Characterization of complete LA-maps. *)
Section Completeness_characterization.

Context {A1 L1 A2 L2}
        (LA1: LabelAlgebra A1 L1)
        (LA2: LabelAlgebra A2 L2)
        (m: LA1 ==> LA2).

Definition completeness_prop :=
  (forall a l1 l2, l1 ⊑[a] l2 -> Lmap m l1 ⊑[Amap m a] Lmap m l2)
  /\ (forall l1 l2, Lmap m l1 =L Lmap m l2 -> l1 =L l2).

Lemma completeness_necessary : LAMapComplete m boolean_sem -> completeness_prop.
Proof.
intro Hcomplete. split.
* intros a l1 l2 H. apply NNPP. intro Hnleq.
  assert (bool_Lequiv (Amap m a) (Lmap m l2)
                      (apply m (Atom (VBool true) l1))
                      (apply m (Atom (VBool false) l1)))
    as HLequiv.
  { intros Hlab; tauto. }
  specialize (Hcomplete _ _ _ _ HLequiv). simpl in Hcomplete.
  specialize (Hcomplete (or_introl H)).
  destruct Hcomplete. congruence.
* intros l1 l2 H.
  assert (bool_Lequiv
            (Amap m bottom) (Lmap m l2)
            (apply m (Atom (VBool true) l1))
            (apply m (Atom (VBool true) l2)))
    as HLequiv.
  { intro Hlab. tauto. }
  specialize (Hcomplete _ _ _ _ HLequiv). simpl in Hcomplete.
  destruct Hcomplete. right; auto. auto.
Qed.

Lemma completeness_sufficient : completeness_prop -> LAMapComplete m boolean_sem.
Proof.
intros [HflowsTo Heq] auth lab [v1 l1] [v2 l2] HLequiv.
destruct v1; destruct v2; simpl in HLequiv; try tauto.
intro Hlab.
assert (Lmap m l1 ⊑[ Amap m auth]Lmap m lab \/
        Lmap m l2 ⊑[ Amap m auth]Lmap m lab) as Hlab'.
{ destruct Hlab; [left | right]; auto. }
specialize (HLequiv Hlab'). destruct HLequiv; auto.
Qed.

Lemma completeness_iff : LAMapComplete m boolean_sem <-> completeness_prop.
Proof.
split.
* apply completeness_necessary.
* apply completeness_sufficient.
Qed.

End Completeness_characterization.

(** * Characterization of embeddings. *)
Section Embedding_characterization.

Context {A1 L1 A2 L2}
        (LA1: LabelAlgebra A1 L1)
        (LA2: LabelAlgebra A2 L2)
        (m: LA1 ==> LA2).

Theorem embedding_iff :
      LAEmbedding m boolean_sem
      <->
      (forall a l1 l2, l1 ⊑[a] l2 <-> Lmap m l1 ⊑[Amap m a] Lmap m l2)
      /\ (forall l1 l2, l1 =L l2 <-> Lmap m l1 =L Lmap m l2).
Proof.
pose proof (soundness_iff m) as [Hsound1 Hsound2].
pose proof (completeness_iff m) as [Hcomplete1 Hcomplete2].
split.
* intros [Hsound Hcomplete].
  specialize (Hsound1 Hsound). destruct Hsound1 as [? ?].
  specialize (Hcomplete1 Hcomplete). destruct Hcomplete1 as [? ?].
  split; firstorder.
* intros [HflowsTo HLeq].
  assert (soundness_prop m).
  { split.
    + intros. rewrite HflowsTo. assumption.
    + intros. rewrite <- HLeq. assumption.
  }
  assert (completeness_prop m).
  { split.
    + intros. rewrite <- HflowsTo. assumption.
    + intros. rewrite HLeq. assumption.
  }
  split; auto.
Qed.

End Embedding_characterization.

(** * Corollaries of the characterization. *)
Section Corollaries.

Context {A1 L1 A2 L2}
        (LA1: LabelAlgebra A1 L1)
        (LA2: LabelAlgebra A2 L2)
        (m: LA1 ==> LA2).

(** Case where the bottom authority is mapped to bottom. *)
Corollary embedding_auth_bottom :
  Amap m bottom =A bottom ->
  (LAEmbedding m boolean_sem
   <->
   (forall a l1 l2, l1 ⊑[a] l2 <-> Lmap m l1 ⊑[Amap m a] Lmap m l2)).
Proof.
intros [Hbottom _]. rewrite embedding_iff.
split; intro H; splits; try tauto.
intros l1 l2. split; intros [H' H''].
* split.
  + apply flowsToBottom.
    apply (flowsToIncluded (Amap m bottom)); auto.
    apply H. auto.
  + apply flowsToBottom.
    apply (flowsToIncluded (Amap m bottom)); auto.
    apply H. auto.
* split.
  + apply flowsToBottom. apply H. auto.
  + apply flowsToBottom. apply H. auto.
Qed.

End Corollaries.

(** * Characterization of boolean-embeddability. *)
Section boolean_embeds.

Context {A1 L1 A2 L2}
        (LA1: LabelAlgebra A1 L1)
        (LA2: LabelAlgebra A2 L2).

Theorem boolean_embeds_iff :
  LA1 ==>[ boolean_sem ] LA2
  <-> exists m : LA1 ==> LA2,
        (forall a l1 l2, l1 ⊑[a] l2 <-> Lmap m l1 ⊑[Amap m a] Lmap m l2)
        /\ (forall l1 l2, l1 =L l2 <-> Lmap m l1 =L Lmap m l2).
Proof.
unfold embeds_in. simpl projT2. unfold embeds_in_open.
split; intros [m Hm]; exists m; apply embedding_iff; assumption.
Qed.

End boolean_embeds.
