Require Import PreLattice.
Require Import Ensembles.
Require Import FinSetLattice.
Require Import Image.

Set Implicit Arguments.

(** * Finite sets or full set, ordered by inclusion. *)
Inductive CFinSet (A: Type) : Type :=
  | CFinSet_intro : FinSet A -> CFinSet A
  | CFinSet_full : CFinSet A.

Definition CFinSet_empty (A: Type) : CFinSet A :=
  CFinSet_intro (FinSet_empty _).

Definition CFinSet_singleton (A: Type) (x: A) : CFinSet A :=
  CFinSet_intro (FinSet_singleton x).

Definition CFinSet_included {A: Type} (s1 s2: CFinSet A) :=
  match s1, s2 with
    | CFinSet_intro s1, CFinSet_intro s2 => FinSet_included s1 s2
    | _, CFinSet_full => True
    | CFinSet_full, _ => False
  end.

Instance CFinSet_included_PreOrder (A: Type) : PreOrder (@CFinSet_included A).
Proof.
unfold CFinSet_included. constructor.
* intros [s |]; trivial. reflexivity.
* intros [s1 |] [s2 |] [s3 |] H12 H23; auto. transitivity s2; trivial.
Qed.

Instance CFinSet_included_HasBottom (A: Type) : HasBottom (CFinSet A) CFinSet_included :=
{ bottom := CFinSet_empty A }.
Proof.
intros [s |]; simpl; auto.
apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
Defined.

Instance CFinSet_included_HasTop (A: Type) : HasTop (CFinSet A) CFinSet_included :=
{ top := CFinSet_full _ }.
Proof.
intros [s |]; simpl; auto.
Defined.

(** * The natural pre-lattice structure on finite sets. *)
Definition CFinSet_union {A: Type} (s1 s2: CFinSet A) : CFinSet A :=
  match s1, s2 with
    | CFinSet_intro s1, CFinSet_intro s2 =>
      CFinSet_intro (FinSet_union s1 s2)
    | _, _ => CFinSet_full _
  end.

Definition CFinSet_add {A: Type} (x: A) (s: CFinSet A) : CFinSet A :=
  CFinSet_union (CFinSet_singleton x) s.

Definition CFinSet_intersection {A: Type} (s1 s2: CFinSet A) : CFinSet A :=
  match s1, s2 with
    | CFinSet_intro s1, CFinSet_intro s2 =>
      CFinSet_intro (FinSet_intersection s1 s2)
    | s, CFinSet_full | CFinSet_full, s => s
  end.

Lemma CFinSet_distrib1 {A: Type} (s1 s2 s3: CFinSet A) :
  CFinSet_included
    (CFinSet_intersection (CFinSet_union s1 s3) (CFinSet_union s2 s3))
    (CFinSet_union (CFinSet_intersection s1 s2) s3).
Proof.
destruct s1 as [s1 |]; destruct s2 as [s2 |]; destruct s3 as [s3 |];
simpl; try reflexivity; auto.
apply FinSet_distrib1.
Qed.

Lemma CFinSet_distrib2 {A: Type} (s1 s2 s3: CFinSet A) :
  CFinSet_included
    (CFinSet_union (CFinSet_intersection s1 s2) s3)
    (CFinSet_intersection (CFinSet_union s1 s3) (CFinSet_union s2 s3)).
Proof.
destruct s1 as [s1 |]; destruct s2 as [s2 |]; destruct s3 as [s3 |];
simpl; try reflexivity; auto.
apply FinSet_distrib2.
Qed.

Instance CFinSet_JoinPreLattice (A: Type) :
  JoinPreLattice (CFinSet A) CFinSet_included :=
{ join := CFinSet_union }.
Proof.
* intros [s1 |] [s2 |]; simpl in *; auto.
  apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
* intros [s1|] [s2|]; simpl in *; auto.
  apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
* intros [s |] [s1 |] [s2 |] H12 H23; simpl in *; auto.
  apply (@join_lub _ _ _ (FinSet_JoinPreLattice _)); trivial.
Defined.

Instance CFinSet_MeetPreLattice (A: Type) :
  MeetPreLattice (CFinSet A) CFinSet_included :=
{ meet := CFinSet_intersection }.
Proof.
* intros [s1 |] [s2 |]; simpl in *; try reflexivity; auto.
  apply (@meet_lb1 _ _ _ (FinSet_MeetPreLattice _)).
* intros [s1|] [s2|]; simpl in *; try reflexivity; auto.
  apply (@meet_lb2 _ _ _ (FinSet_MeetPreLattice _)).
* intros [s |] [s1 |] [s2 |] H12 H23; simpl in *; auto.
  apply (@meet_glb _ _ _ (FinSet_MeetPreLattice _)); trivial.
Defined.

(** ** Union with empty set. *)
Lemma CFinSet_union_with_empty_set_included1 {A: Type} (s1 s2: CFinSet A) :
  CFinSet_included s1 (CFinSet_union s2 (CFinSet_empty _))
  <-> CFinSet_included s1 s2.
Proof.
destruct s1; destruct s2; simpl; try tauto.
apply FinSet_union_with_empty_set_included1.
Qed.

Lemma CFinSet_union_with_empty_set_included2 {A: Type} (s1 s2: CFinSet A) :
  CFinSet_included (CFinSet_union s1 (CFinSet_empty _)) s2
  <-> CFinSet_included s1 s2.
Proof.
destruct s1; destruct s2; simpl; try tauto.
apply FinSet_union_with_empty_set_included2.
Qed.

(** * Image of finite sets by a function. *)
Definition CFinSet_im {A B: Type} (X: CFinSet A) (f:A -> B) : CFinSet B :=
  match X with
    | CFinSet_intro X => CFinSet_intro (FinSet_im X f)
    | CFinSet_full => CFinSet_full _
  end.

Lemma CFinSet_im_union {A B : Type} :
  forall (X Y: CFinSet A) (f: A -> B),
    equiv (CFinSet_im (CFinSet_union X Y) f)
          (CFinSet_union (CFinSet_im X f) (CFinSet_im Y f)).
Proof.
intros [X|] [Y|] f; simpl; auto. apply FinSet_im_union.
Qed.

Lemma CFinSet_im_intersection {A B : Type} :
  forall (X Y: CFinSet A) (f: A -> B),
    injective _ _ f ->
    equiv (CFinSet_im (CFinSet_intersection X Y) f)
          (CFinSet_intersection (CFinSet_im X f) (CFinSet_im Y f)).
Proof.
intros [X|] [Y|] f Hinj; try reflexivity; auto.
apply FinSet_im_intersection; trivial.
Qed.

Lemma CFinSet_im_singleton {A B: Type} :
  forall (f: A -> B) (x: A),
    equiv (CFinSet_im (CFinSet_singleton x) f)
          (CFinSet_singleton (f x)).
Proof.
intros f x. apply FinSet_im_singleton.
Qed.

Lemma CFinSet_im_add {A B: Type} :
  forall (X:CFinSet A) (x:A) (f:A -> B),
    equiv (CFinSet_im (CFinSet_add x X) f)
          (CFinSet_add (f x) (CFinSet_im X f)).
Proof.
intros [X|] x f.
* apply FinSet_im_add.
* reflexivity.
Qed.

Lemma CFinSet_im_empty {A B: Type} (f: A -> B) :
  equiv (CFinSet_im (CFinSet_empty A) f) (CFinSet_empty B).
Proof.
apply FinSet_im_empty.
Qed.
