Require Import PreLattice.
Require Import Ensembles.
Require Import SetLattice.
Require Import SetFacts.
Require Import Finite_sets_facts.
Require Import Image.

Set Implicit Arguments.

(** * Finite sets ordered by inclusion. *)
Definition FinSet (A: Type) := { S: Ensemble A | Finite A S }.

Definition FinSet_empty (A: Type) : FinSet A :=
  exist _ (Empty_set A) (Empty_is_finite A).

Definition FinSet_singleton (A: Type) (x: A) : FinSet A :=
  exist _ (Add A (Empty_set A) x)
        (Add_preserves_Finite _ (Empty_set A) x (Empty_is_finite A)).

Definition FinSet_included {A: Type} (s1 s2: FinSet A) :=
  Included _ (proj1_sig s1) (proj1_sig s2).

Instance FinSet_included_PreOrder (A: Type) : PreOrder (@FinSet_included A).
Proof.
unfold FinSet_included. constructor; auto with sets.
Qed.

Instance FinSet_included_HasBottom (A: Type) : HasBottom (FinSet A) FinSet_included :=
{ bottom := FinSet_empty A }.
Proof.
intros [s H]. unfold FinSet_empty. unfold FinSet_included.
simpl. auto with sets.
Defined.

(** * The natural pre-lattice structure on finite sets. *)
Definition FinSet_union {A: Type} (s1 s2: FinSet A) : FinSet A :=
exist _
  (Union _ (proj1_sig s1) (proj1_sig s2))
  (Union_preserves_Finite _
     (proj1_sig s1) (proj1_sig s2)
     (proj2_sig s1) (proj2_sig s2)).

Definition FinSet_add {A: Type} (x: A) (s: FinSet A) : FinSet A :=
  FinSet_union (FinSet_singleton x) s.

Definition FinSet_intersection {A: Type} (s1 s2: FinSet A) : FinSet A :=
exist _
  (Intersection _ (proj1_sig s1) (proj1_sig s2))
  (Intersection_preserves_finite _
     (proj1_sig s2) (proj2_sig s2) (proj1_sig s1)).

Lemma FinSet_distrib1 {A: Type} (s1 s2 s3: FinSet A) :
  FinSet_included
    (FinSet_intersection (FinSet_union s1 s3) (FinSet_union s2 s3))
    (FinSet_union (FinSet_intersection s1 s2) s3).
Proof.
destruct s1 as [s1 Hs1]. destruct s2 as [s2 Hs2]. destruct s3 as [s3 Hs3].
unfold FinSet_union in *. unfold FinSet_intersection in *.
unfold FinSet_included in *.
simpl in *.
apply Ensemble_distrib1.
Qed.

Lemma FinSet_distrib2 {A: Type} (s1 s2 s3: FinSet A) :
  FinSet_included
    (FinSet_union (FinSet_intersection s1 s2) s3)
    (FinSet_intersection (FinSet_union s1 s3) (FinSet_union s2 s3)).
Proof.
destruct s1 as [s1 Hs1]. destruct s2 as [s2 Hs2]. destruct s3 as [s3 Hs3].
unfold FinSet_union in *. unfold FinSet_intersection in *.
unfold FinSet_included in *.
simpl in *. apply Ensemble_distrib2.
Qed.

Instance FinSet_JoinPreLattice (A: Type) :
  JoinPreLattice (FinSet A) FinSet_included :=
{ join := FinSet_union }.
Proof.
* intros [s1 H1] [s2 H2] p Hp. simpl in *. auto with sets.
* intros [s1 H1] [s2 H2] p Hp. simpl in *. auto with sets.
* intros [s H] [s1 H1] [s2 H2] p Hp.
  unfold FinSet_included in *.
  simpl in *. auto with sets.
Defined.

Instance FinSet_MeetPreLattice (A: Type) :
  MeetPreLattice (FinSet A) FinSet_included :=
{ meet := FinSet_intersection }.
Proof.
* intros [s1 H1] [s2 H2] p Hp. simpl in *.
  apply Intersection_inv in Hp. tauto.
* intros [s1 H1] [s2 H2] p Hp. simpl in *.
  apply Intersection_inv in Hp. tauto.
* intros [s H] [s1 H1] [s2 H2] p Hp.
  unfold FinSet_included in *.
  simpl in *. auto with sets.
Defined.

(** * About [FinSet_included]. *)
Lemma FinSet_In {A} (X Y: FinSet A) (x: A) :
  FinSet_included X Y ->
  In _ (proj1_sig X) x ->
  In _ (proj1_sig Y) x.
Proof.
intros H HIn. destruct X as [X HX]. destruct Y as [Y HY].
unfold FinSet_included in H. simpl in *. auto with sets.
Qed.

(** ** Union with empty set. *)
Lemma FinSet_union_with_empty_set_included1 {A: Type} (s1 s2: FinSet A) :
  FinSet_included s1 (FinSet_union s2 (FinSet_empty _))
  <-> FinSet_included s1 s2.
Proof.
apply Union_with_empty_set_Included1.
Qed.

Lemma FinSet_union_with_empty_set_included2 {A: Type} (s1 s2: FinSet A) :
  FinSet_included (FinSet_union s1 (FinSet_empty _)) s2
  <-> FinSet_included s1 s2.
Proof.
apply Union_with_empty_set_Included2.
Qed.

(** * Image of finite sets by a function. *)
Definition FinSet_im {A B: Type} (X: FinSet A) (f:A -> B) : FinSet B :=
  exist _
        (Im _ _ (proj1_sig X) f)
        (finite_image _ _ (proj1_sig X) f (proj2_sig X)).

Lemma FinSet_im_def {A B: Type} :
  forall (X:FinSet A) (f:A -> B) (x:A),
    In _ (proj1_sig X) x ->
    In _ (proj1_sig (FinSet_im X f)) (f x).
Proof.
intros X f x H. unfold FinSet_im. simpl.
auto with sets.
Qed.

Lemma FinSet_im_union {A B : Type} :
  forall (X Y: FinSet A) (f: A -> B),
    equiv (FinSet_im (FinSet_union X Y) f)
          (FinSet_union (FinSet_im X f) (FinSet_im Y f)).
Proof.
intros X Y f.
unfold FinSet_im, FinSet_union. simpl.
unfold FinSet_included. simpl. split; intros b H.
* apply Im_inv in H. destruct H as [a [H ?]]. subst b.
  apply Union_inv in H. destruct H as [H | H]; auto with sets.
* apply Union_inv in H.
  destruct H as [H | H]; apply Im_inv in H;
  destruct H as [a [H ?]]; subst b; auto with sets.
Qed.

Lemma Im_intersection {A B : Type} :
  forall (X Y: Ensemble A) (f: A -> B),
    injective _ _ f ->
    Im _ _ (Intersection _ X Y) f = Intersection _ (Im _ _ X f) (Im _ _ Y f).
Proof.
intros X Y f Hinj. apply Extensionality_Ensembles.
split; intros b H.
* apply Im_inv in H. destruct H as [a [H ?]]. subst b.
  apply Intersection_inv in H. destruct H. auto with sets.
* apply Intersection_inv in H. destruct H as [HX HY].
  apply Im_inv in HX. destruct HX as [aX [HX ?]]. subst b.
  apply Im_inv in HY. destruct HY as [aY [HY Heq]].
  apply Hinj in Heq. subst aY.
  auto with sets.
Qed.
Hint Resolve Im_intersection : sets.

Lemma FinSet_im_intersection {A B : Type} :
  forall (X Y: FinSet A) (f: A -> B),
    injective _ _ f ->
    equiv (FinSet_im (FinSet_intersection X Y) f)
          (FinSet_intersection (FinSet_im X f) (FinSet_im Y f)).
Proof.
intros X Y f Hinj.
unfold FinSet_im, FinSet_intersection. simpl.
unfold FinSet_included. simpl.
rewrite Im_intersection; trivial. split; reflexivity.
Qed.

Lemma FinSet_im_singleton {A B: Type} :
  forall (f: A -> B) (x: A),
    equiv (FinSet_im (FinSet_singleton x) f)
          (FinSet_singleton (f x)).
Proof.
intros f x. unfold FinSet_im, FinSet_singleton. simpl.
unfold FinSet_included. simpl. split; intros b H.
* apply Im_inv in H. destruct H as [a [H ?]]. subst b.
  apply Union_inv in H. destruct H as [[] | H].
  apply Singleton_inv in H. subst x. auto with sets.
* apply Union_inv in H. destruct H as [[] | H].
  apply Singleton_inv in H. subst b. auto with sets.
Qed.

Lemma FinSet_im_add {A B: Type} :
  forall (X:FinSet A) (x:A) (f:A -> B),
    equiv (FinSet_im (FinSet_add x X) f)
          (FinSet_add (f x) (FinSet_im X f)).
Proof.
intros X x f.
unfold FinSet_im, FinSet_add, FinSet_union, FinSet_singleton, FinSet_included.
simpl. split; intros b H.
* apply Im_inv in H. destruct H as [a [H ?]]. subst b.
  apply Union_inv in H. destruct H as [H | H].
  + left. apply Union_inv in H. destruct H as [[] | H].
    apply Singleton_inv in H. subst x. auto with sets.
  + auto with sets.
* apply Union_inv in H. destruct H as [H | H].
  + apply Union_inv in H. destruct H as [[] | H].
    apply Singleton_inv in H. subst b. auto with sets.
  + apply Im_inv in H. destruct H as [y [H ?]]. subst b.
    auto with sets.
Qed.

Lemma FinSet_im_empty {A B: Type} (f: A -> B) :
  equiv (FinSet_im (FinSet_empty A) f) (FinSet_empty B).
Proof.
unfold FinSet_empty. unfold FinSet_im. simpl. unfold FinSet_included. simpl.
split; intros b H.
* apply Im_inv in H. destruct H as [? [[] ?]].
* destruct H.
Qed.
