Require Import MyTactics.
Require Import LabelAlgebra.
Require Import RelLattice.
Require Import MyClassical.

Set Implicit Arguments.

(** * Label algebra semantics. *)
Section Sem.

Context {F: Type -> Type -> Type} {M: LAMapApply F}.

Definition SEM := forall (A L: Type) (LA: LabelAlgebra A L),
                    A -> L -> F A L -> F A L -> Prop.

Definition good_LASemantics (sem: SEM) :=
  forall A L (LA: LabelAlgebra A L),
    (forall (a a': A) (l: L),
       a ≤ a' ->
       coarser (sem _ _ LA a l) (sem _ _ LA a' l))
    /\ (forall (a: A) (l l': L),
          l ⊑[a] l' ->
          coarser (sem _ _ LA a l) (sem _ _ LA a l'))
    /\ (forall A' L' (LA': LabelAlgebra A' L') (m: LA ==> LA'),
          LAinjection m ->
          forall a l x1 x2,
            sem _ _ LA a l x1 x2 <->
            sem _ _ LA' (Amap m a) (Lmap m l) (apply m x1) (apply m x2)).

End Sem.

Arguments SEM: clear implicits.

(** * Sound and complete maps; embeddings. *)
Section Embeddings.

Context {A1 L1 A2 L2: Type}
        {LA1: LabelAlgebra A1 L1} {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2).

Context {F: Type -> Type -> Type} {M: LAMapApply F}
        (sem: SEM F).

Definition LAMapSound :=
  forall a l x1 x2,
    sem LA1 a l x1 x2 ->
    sem LA2 (Amap m a) (Lmap m l) (apply m x1) (apply m x2).

Definition LAMapComplete :=
  forall a l x1 x2,
    sem LA2 (Amap m a) (Lmap m l) (apply m x1) (apply m x2) ->
    sem LA1 a l x1 x2.

Definition LAEmbedding := LAMapSound /\ LAMapComplete.

Lemma LAEmbedding_rewrite :
  LAEmbedding <->
  forall a l x1 x2,
    sem LA1 a l x1 x2 <->
    sem LA2 (Amap m a) (Lmap m l) (apply m x1) (apply m x2).
Proof.
split; intro H.
* destruct H as [Hsound Hcomplete].
  intros a l x1 x2.
  specialize (Hsound a l x1 x2).
  specialize (Hcomplete a l x1 x2).
  tauto.
* split; intros a l x1 x2; specialize (H a l x1 x2); tauto.
Qed.

End Embeddings.

(** * General properties about embeddings. *)
Section Facts.

Context {A1 L1 A2 L2 A3 L3: Type}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        {LA3: LabelAlgebra A3 L3}
        (m12: LA1 ==> LA2) (m23: LA2 ==> LA3).

Context {F: Type -> Type -> Type} {M: LAMapApply F} (sem: SEM F).

Lemma sound_LAmap_refl : LAMapSound (LAidentity LA1) sem.
Proof.
intros a l x1 x2. simpl.
rewrite apply_refl. rewrite apply_refl.
trivial.
Qed.

Lemma complete_LAmap_refl : LAMapComplete (LAidentity LA1) sem.
Proof.
intros a l x1 x2. simpl.
rewrite apply_refl. rewrite apply_refl.
trivial.
Qed.

(** [LAidentity] is an endo-embedding. *)
Lemma embedding_refl : LAEmbedding (LAidentity LA1) sem.
Proof.
split. apply sound_LAmap_refl. apply complete_LAmap_refl.
Qed.

Lemma sound_LAmap_trans :
  LAMapSound m12 sem -> LAMapSound m23 sem ->
  LAMapSound (LAcompose m23 m12) sem.
Proof.
intros H12 H23 a l x1 x1'. simpl.
rewrite apply_compose. rewrite apply_compose.
auto.
Qed.

Lemma complete_LAmap_trans :
  LAMapComplete m12 sem -> LAMapComplete m23 sem ->
  LAMapComplete (LAcompose m23 m12) sem.
Proof.
intros H12 H23 a l x1 x1'. simpl.
rewrite apply_compose. rewrite apply_compose.
auto.
Qed.

(** Embeddings are compatible with composition. *)
Lemma embedding_trans :
  LAEmbedding m12 sem -> LAEmbedding m23 sem ->
  LAEmbedding (LAcompose m23 m12) sem.
Proof.
intros [? ?] [? ?]. split; auto using sound_LAmap_trans, complete_LAmap_trans.
Qed.

End Facts.

(** * Embeddability (open version) *)
Section Embeddability_open.

Context {F: Type -> Type -> Type}
        {M: LAMapApply F}.

(** Embeddability is the existence of an embedding. *)
Definition embeds_in_open {A1 L1 A2 L2: Type}
           (sem: SEM F)
           (LA1: LabelAlgebra A1 L1)
           (LA2: LabelAlgebra A2 L2) :=
  exists m : LA1 ==> LA2,
    LAEmbedding m sem.

(** Embeddability is reflexive and transitive. *)
Lemma embeds_in_open_refl :
  forall {A L} (sem: SEM F) (LA: LabelAlgebra A L), embeds_in_open sem LA LA.
Proof.
intros A L sem LA. exists (LAidentity LA).
apply embedding_refl.
Qed.

Lemma embeds_in_open_trans :
  forall {A1 L1 A2 L2 A3 L3}
         (sem: SEM F)
         (LA1: LabelAlgebra A1 L1)
         (LA2: LabelAlgebra A2 L2)
         (LA3: LabelAlgebra A3 L3),
    embeds_in_open sem LA1 LA2 ->
    embeds_in_open sem LA2 LA3 ->
    embeds_in_open sem LA1 LA3.
Proof.
intros A1 L1 A2 L2 A3 L3 sem LA1 LA2 LA3 [m12 H12] [m23 H23].
exists (LAcompose m23 m12).
apply embedding_trans; assumption.
Qed.

End Embeddability_open.

(** * Embeddability (packed version) *)
Section Embeddability.

Context {F: Type -> Type -> Type}
        {M: LAMapApply F}.

(** Embeddability is the existence of an embedding. *)
Definition embeds_in
           (sem: SEM F)
           (LA1: { A1 : Type & { L1 : Type & LabelAlgebra A1 L1 } })
           (LA2: { A2 : Type & { L2 : Type & LabelAlgebra A2 L2 } }) :=
   embeds_in_open sem (projT2 (projT2 LA1)) (projT2 (projT2 LA2)).

(** Embeddability is reflexive and transitive. *)
Lemma embeds_in_refl :
  forall (sem: SEM F) (LA: {A:Type & {L:Type & LabelAlgebra A L}}),
    embeds_in sem LA LA.
Proof.
intros sem [A [L LA]]. unfold embeds_in. simpl.
apply embeds_in_open_refl.
Qed.

Lemma embeds_in_trans :
  forall (sem: SEM F)
         (LA1: {A:Type & {L:Type & LabelAlgebra A L}})
         (LA2: {A:Type & {L:Type & LabelAlgebra A L}})
         (LA3: {A:Type & {L:Type & LabelAlgebra A L}}),
    embeds_in sem LA1 LA2 ->
    embeds_in sem LA2 LA3 ->
    embeds_in sem LA1 LA3.
Proof.
intros sem [A1 [L1 LA1]] [A2 [L2 LA2]] [A3 [L3 LA3]] H12 H23.
unfold embeds_in in *. simpl in *.
apply @embeds_in_open_trans with (LA2 := LA2); assumption.
Qed.

Global Instance embeds_in_Preorder (sem: SEM F) : PreOrder (embeds_in sem).
Proof.
constructor.
* intros LA. apply embeds_in_refl.
* intros LA1 LA2 LA3. apply embeds_in_trans.
Qed.

End Embeddability.

(** Notation for embeddability between already packed label algebras. *)
Notation "LA1 #==>[ sem ] LA2" := (@embeds_in _ _ sem LA1 LA2) (at level 70).

(** Notation for embeddability between label algebras that are not packed. *)
Notation "LA1 ==>[ sem ] LA2" :=
  (@embeds_in _ _ sem (packLA LA1) (packLA LA2)) (at level 70).

(** Useful tactics to prove embeddability. *)
Hint Extern 1 (_ #==>[ _ ] _) => reflexivity.
Hint Extern 1 (_ ==>[ _ ] _) => reflexivity.

(** Short-hand tactic to allow easy transitive reasoning. *)
Ltac transitivity_embeds LA := transitivity (packLA LA).

Lemma embeds_in_trans_not1
      {F: Type -> Type -> Type} {M: LAMapApply F}
      (sem: SEM F)
      {A1 A2 A3 L1 L2 L3}
      (LA1: LabelAlgebra A1 L1)
      (LA2: LabelAlgebra A2 L2)
      (LA3: LabelAlgebra A3 L3) :
  LA1 ==>[sem] LA2 ->
  ~(LA1 ==>[sem] LA3) ->
  ~(LA2 ==>[sem] LA3).
Proof.
intro H. apply contraposition. intro H'.
transitivity_embeds LA2; assumption.
Qed.

Lemma embeds_in_trans_not2
      {F: Type -> Type -> Type} {M: LAMapApply F}
      (sem: SEM F)
      {A1 A2 A3 L1 L2 L3}
      (LA1: LabelAlgebra A1 L1)
      (LA2: LabelAlgebra A2 L2)
      (LA3: LabelAlgebra A3 L3) :
  LA2 ==>[sem] LA3 ->
  ~(LA1 ==>[sem] LA3) ->
  ~(LA1 ==>[sem] LA2).
Proof.
intro H. apply contraposition. intro H'.
transitivity_embeds LA2; assumption.
Qed.

(** * Embeddings and injections. *)

(** A simple semantics of label algebras, for which embeddings are injections. *)
Section Slab_def.

Context (A L: Type) (LA: LabelAlgebra A L).

Inductive Slab : Type :=
| Slab_nop : Slab
| Slab_AeqBottom : A -> Slab
| Slab_Aleq : A -> A -> Slab
| Slab_Ajoin : A -> A -> A -> Slab
| Slab_LeqLdef : L -> Slab
| Slab_flowsTo : A -> L -> L -> Slab
| Slab_Ljoin : L -> L -> L -> Slab
| Slab_Lmeet : L -> L -> L -> Slab.

Definition Slab_interp (s: Slab) : Prop :=
  match s with
    | Slab_nop => True
    | Slab_AeqBottom a => a =A bottom
    | Slab_Aleq a1 a2 => a1 ≤ a2
    | Slab_Ajoin a1 a2 a3 => a1 ∨ a2 =A a3
    | Slab_LeqLdef l => l =L Ldef
    | Slab_flowsTo a l1 l2 => l1 ⊑[a] l2
    | Slab_Ljoin l1 l2 l3 => l1 ⊔ l2 =L l3
    | Slab_Lmeet l1 l2 l3 => l1 ⊓ l2 =L l3
  end.

End Slab_def.

Definition Slab_apply A1 L1 (LA1: LabelAlgebra A1 L1)
        A2 L2 (LA2: LabelAlgebra A2 L2)
        (m: LA1 ==> LA2) (s: Slab A1 L1) : Slab A2 L2 :=
    match s with
      | Slab_nop => Slab_nop _ _
      | Slab_AeqBottom a =>
          Slab_AeqBottom _ (Amap m a)
      | Slab_Aleq a1 a2 =>
          Slab_Aleq _ (Amap m a1) (Amap m a2)
      | Slab_Ajoin a1 a2 a3 =>
          Slab_Ajoin _ (Amap m a1) (Amap m a2) (Amap m a3)
      | Slab_LeqLdef l =>
          Slab_LeqLdef _ (Lmap m l)
      | Slab_flowsTo a l1 l2 =>
          Slab_flowsTo (Amap m a) (Lmap m l1) (Lmap m l2)
      | Slab_Ljoin l1 l2 l3 =>
          Slab_Ljoin _ (Lmap m l1) (Lmap m l2) (Lmap m l3)
      | Slab_Lmeet l1 l2 l3 =>
          Slab_Lmeet _ (Lmap m l1) (Lmap m l2) (Lmap m l3)
    end.

Definition Slab_sem : SEM Slab :=
  fun A L (LA: LabelAlgebra A L) a l _ => Slab_interp LA.

Instance LAMapApplySlab : LAMapApply Slab :=
{ apply := Slab_apply }.
Proof.
* intros A L LA x; destruct x; reflexivity.
* intros A L LA A' L' LA' A'' L'' LA'' m1 m2 x; destruct x; reflexivity.
Defined.

Lemma Slab_good_LASemantics : good_LASemantics Slab_sem.
Proof.
intros A L LA. splits.
* intros a a' l Ha s1 s2 H. assumption.
* intros a l l' Hl s1 s2 H. assumption.
* intros A' L' LA' m' Hinj a l s1 s2.
  destruct Hinj as
      [[HAcompat [HAbottom [HAjoin [HLcompat [HLdef [HLjoin [HLmeet Hmon]]]]]]]
         [[HAinj HLinj] HflowsTo]].
  unfold Slab_sem.
  destruct s2; unfold apply, LAMapApplySlab, Slab_apply, Slab_interp; auto.
  + tauto.
  + split; intro H.
    - transitivity (Amap m' bottom); auto.
    - apply HAinj. rewrite HAbottom. assumption.
  + split; intro H.
    - assert (a0 ∨ a1 =A a1) as Ha. { split; auto. }
      apply HAcompat in Ha. rewrite <- HAjoin in Ha.
      apply join_characterization. rewrite Ha. reflexivity.
    - assert (a0 ∨ a1 =A a1) as Ha.
      { apply HAinj. rewrite <- HAjoin. split; auto. }
      apply join_characterization. rewrite Ha. reflexivity.
  + split; intro H.
    - transitivity (Amap m' (a0 ∨ a1)); auto.
    - specialize (HAjoin a0 a1).
      split; apply HAinj; rewrite <- H; rewrite HAjoin; reflexivity.
  + split; intro H.
    - transitivity (Lmap m' Ldef); auto.
    - split; apply HLinj; rewrite HLdef; auto. symmetry; assumption.
  + split; intro H; auto.
  + split; intro H.
    - transitivity (Lmap m' (l0 ⊔ l1)); auto.
    - specialize (HLjoin l0 l1).
      split; apply HLinj; rewrite <- H; auto. symmetry; auto.
  + split; intro H.
    - transitivity (Lmap m' (l0 ⊓ l1)); auto.
    - specialize (HLmeet l0 l1).
      split; apply HLinj; rewrite <- H; auto. symmetry; auto.
Qed.

Section Injection.

Context {A1 L1 A2 L2 A3 L3: Type}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2).

Context {F: Type -> Type -> Type} {M: LAMapApply F} (sem: SEM F)
        {Hgood: good_LASemantics sem}.

Lemma LAinjection_embedding : LAinjection m -> LAEmbedding m sem.
Proof.
intro Hiso. apply LAEmbedding_rewrite.
destruct (Hgood LA1) as [_ [_ Hsemiso]]. auto.
Qed.

Lemma Slab_sem_embedding : LAEmbedding m Slab_sem -> LAinjection m.
Proof.
intro H. rewrite (LAEmbedding_rewrite m Slab_sem) in H.
specialize (H bottom Ldef (Slab_nop _ _)). unfold Slab_sem in H.
apply LAinjection_iff. unfold LAinjection_prop. splits.
* specialize (H (Slab_AeqBottom _ bottom)). apply H.
  unfold Slab_interp. reflexivity.
* intros a1 a2. specialize (H (Slab_Aleq _ a1 a2)). assumption.
* intros a1 a2. specialize (H (Slab_Ajoin _ a1 a2 (a1 ∨ a2))).
  apply H. unfold Slab_interp. reflexivity.
* specialize (H (Slab_LeqLdef _ Ldef)). apply H.
  unfold Slab_interp. reflexivity.
* intros l1 l2. specialize (H (Slab_Ljoin _ l1 l2 (l1 ⊔ l2))).
  apply H. unfold Slab_interp. reflexivity.
* intros l1 l2. specialize (H (Slab_Lmeet _ l1 l2 (l1 ⊓ l2))).
  apply H. unfold Slab_interp. reflexivity.
* intros a l1 l2. specialize (H (Slab_flowsTo a l1 l2)). assumption.
Qed.

End Injection.

Lemma LAEmbedding_injection
      {A1 A2 L1 L2}
      (LA1: LabelAlgebra A1 L1)
      (LA2: LabelAlgebra A2 L2) :
  (forall (F: Type -> Type -> Type) (M: LAMapApply F) (sem: SEM F),
     good_LASemantics sem -> LA1 ==>[sem] LA2)
  <-> LA1 ↪ LA2.
Proof.
split.
* intro H. specialize (H Slab LAMapApplySlab Slab_sem Slab_good_LASemantics).
  destruct H as [m H].
  exists m. apply Slab_sem_embedding. assumption.
* intros [m H] F M sem. exists m. apply LAinjection_embedding; auto.
Qed.

Lemma LAEmbedding_injection_contra
      {A1 A2 L1 L2}
      (LA1: LabelAlgebra A1 L1)
      (LA2: LabelAlgebra A2 L2)
      (F: Type -> Type -> Type)
      (M: LAMapApply F)
      (sem: SEM F) :
  good_LASemantics sem ->
  ~(LA1 ==>[sem] LA2) ->
  ~(LA1 ↪ LA2).
Proof.
intro H0. apply contraposition.
intro H1. apply LAEmbedding_injection; assumption.
Qed.
