Require Import MyTactics.
Require Import LabelAlgebra.
Require Import Lambda.
Require Import Auth.

(** * Authority-wise comparison of terms. *)
Section Term.

Context {A L} {LA: LabelAlgebra A L}.

Fixpoint term_auth_leq (t1 t2: term A L) {struct t1} : Prop :=
match t1, t2 with
| TBool b1, TBool b2 => b1 = b2
| TVar n1, TVar n2 => n1 = n2
| TLam t1, TLam t2 => term_auth_leq t1 t2
| TApp t1 t1', TApp t2 t2' => term_auth_leq t1 t2 /\ term_auth_leq t1' t2'
| TRelabel t1 a1 l1, TRelabel t2 a2 l2 =>
  term_auth_leq t1 t2 /\ a1 ≤ a2 /\ l1 = l2
| TBool _, _ | TVar _, _ | TLam _, _
| TApp _ _, _ | TRelabel _ _ _, _ => False
end.

Global Instance Reflexive_term_auth_leq : Reflexive term_auth_leq.
Proof.
intro t. induction t; simpl; auto.
Qed.

Global Instance Transitive_term_auth_leq : Transitive term_auth_leq.
Proof.
intros t1.
induction t1; intros t2 t3 H12 H23;
destruct t2; destruct t3; simpl in *; try tauto; eauto.
* intuition eauto.
* intuition.
  + eauto.
  + transitivity a0; trivial.
  + congruence.
Qed.

End Term.

Hint Extern 1 =>
match goal with
| |- term_auth_leq _ _ => reflexivity
end.

(** * Authority-wise comparison of atoms, values, environments. *)
Section Atom.

Context {A L} {LA: LabelAlgebra A L}.

Fixpoint atom_auth_leq (a1 a2: atom A L) {struct a1} : Prop :=
match a1, a2 with
| Atom v1 l1, Atom v2 l2 =>
  value_auth_leq v1 v2 /\ l1 = l2
end
with value_auth_leq (v1 v2: value A L) {struct v1} : Prop :=
match v1, v2 with
| VBool b1, VBool b2 => b1 = b2
| VClos e1 t1, VClos e2 t2 =>
  list_forall2 atom_auth_leq e1 e2
  /\ term_auth_leq t1 t2
| VBool _, _ | VClos _ _, _ => False
end.

Definition env_auth_leq := list_forall2 atom_auth_leq.

Lemma atom_value_env_auth_leq_refl :
  (forall a, atom_auth_leq a a)
  /\ (forall v, value_auth_leq v v)
  /\ (forall e, env_auth_leq e e).
Proof.
apply atom_value_env_mutind; intros; simpl; auto.
induction l; simpl; auto. split.
- apply (H 0 a). reflexivity.
- apply IHl. intros n b Hb. apply (H (S n) b). assumption.
Qed.

Lemma atom_value_env_auth_leq_trans :
  (forall a1 a2 a3,
     atom_auth_leq a1 a2 -> atom_auth_leq a2 a3 -> atom_auth_leq a1 a3)
  /\ (forall v1 v2 v3,
        value_auth_leq v1 v2 -> value_auth_leq v2 v3 -> value_auth_leq v1 v3)
  /\ (forall e1 e2 e3,
        env_auth_leq e1 e2 -> env_auth_leq e2 e3 -> env_auth_leq e1 e3).
Proof.
apply atom_value_env_mutind; intros; simpl in *.
* destruct a2. simpl in *. destruct a3. intuition.
  eauto. etransitivity; eassumption.
* destruct v2; simpl in *; destruct v3; eauto.
* destruct v2; simpl in *; destruct v3; intuition auto.
  eapply H; eassumption.
  etransitivity; eassumption.
* generalize dependent e3. generalize dependent e2.
  induction l; intros e2 H12 e3 H23; destruct e2; simpl in *;
  destruct e3; try tauto.
  split.
  - apply (H 0 a eq_refl a0); tauto.
  - apply IHl with (e2 := e2); try tauto.
    intros n b Hb. apply (H (S n) b). assumption.
Qed.

Global Instance Reflexive_atom_auth_leq : Reflexive atom_auth_leq.
Proof.
destruct atom_value_env_auth_leq_refl as [? _]. assumption.
Qed.

Global Instance Transitive_atom_auth_leq : Transitive atom_auth_leq.
Proof.
destruct atom_value_env_auth_leq_trans as [? _]. assumption.
Qed.

Global Instance Reflexive_value_auth_leq : Reflexive value_auth_leq.
Proof.
destruct atom_value_env_auth_leq_refl as [_ [? _]]. assumption.
Qed.

Global Instance Transitive_value_auth_leq : Transitive value_auth_leq.
Proof.
destruct atom_value_env_auth_leq_trans as [_ [? _]]. assumption.
Qed.

Global Instance Reflexive_env_auth_leq : Reflexive env_auth_leq.
Proof.
destruct atom_value_env_auth_leq_refl as [_ [_ ?]]. assumption.
Qed.

Global Instance Transitive_env_auth_leq : Transitive env_auth_leq.
Proof.
destruct atom_value_env_auth_leq_trans as [_ [_ ?]]. assumption.
Qed.

End Atom.

Hint Unfold atom_auth_leq value_auth_leq env_auth_leq.

Hint Extern 1 =>
match goal with
| |- atom_auth_leq _ _ => reflexivity
| |- value_auth_leq _ _ => reflexivity
| |- env_auth_leq _ _ => reflexivity
end.

(** * Authority-wise comparison is consistent with authority of terms, atoms, values, environments. *)
Section Auth.

Context {A L} {LA: LabelAlgebra A L}.

Lemma term_auth_leq_consistent :
  forall (t1 t2: term A L),
    term_auth_leq t1 t2 ->
    term_auth t1 ≤ term_auth t2.
Proof.
intros t1. induction t1; intros t2 H; destruct t2; simpl in *; try tauto; auto.
* destruct H. apply join_lub.
  transitivity (term_auth t2_1); auto.
  transitivity (term_auth t2_2); auto.
* destruct H as [? [Ha ?]].
  rewrite Ha. rewrite IHt1; auto.
Qed.

End Auth.

(** * "Least privilege": running a program with a greater authority makes no harm. *)
Section Least.

Context {A L} {LA: LabelAlgebra A L}.

Lemma increase_auth_strong :
  forall pc e1 e2 t1 t2 a1,
    pc; e1 ⊢ t1 ⇓ a1 ->
    env_auth_leq e1 e2 ->
    term_auth_leq t1 t2 ->
    exists a2,
      pc; e2 ⊢ t2 ⇓ a2 /\ atom_auth_leq a1 a2.
Proof.
intros pc e1 e2 t1 t2 a1 Heval.
generalize dependent t2. generalize dependent e2.
induction Heval; intros e2 t2' Henv Ht; destruct t2'; simpl in *; try tauto.
* subst b0. eauto.
* subst n0.
  destruct (list_forall2_atIndex_fun _ _ _ _ _ Henv H) as [[v2 l2] [Hn2 Ha]].
  destruct Ha as [Hv Hl]. fold (value_auth_leq v v2) in Hv. subst l2.
  exists (Atom v2 (l ⊔ pc)). auto.
* exists (Atom (VClos e2 t2') pc). auto.
* destruct Ht as [Ht1 Ht2].
  specialize (IHHeval1 _ _ Henv Ht1).
  destruct IHHeval1 as [[v1' l1'] [Heval1' Hatom1']].
  destruct v1'; red in Hatom1'; try tauto.
  fold atom_auth_leq in Hatom1'. fold (env_auth_leq e1' l) in Hatom1'.
  rename l into e2'. rename t into t2'.
  destruct Hatom1' as [[Henv' Ht'] ?]. subst l1'.
  specialize (IHHeval2 _ _ Henv Ht2).
  destruct IHHeval2 as [a2' [Heval2' Hatom2']].
  specialize (IHHeval3 (a2' :: e2') _ (conj Hatom2' Henv') Ht').
  destruct IHHeval3 as [a3' [Heval3' Hatom3']].
  eauto.
* destruct Ht as [Ht [Ha Hl]]. subst l0.
  specialize (IHHeval _ _ Henv Ht).
  destruct IHHeval as [[v2 l2] [Heval2 Hatom2]].
  destruct Hatom2 as [Hv Hl].
  fold value_auth_leq in Hv. subst l2.
  exists (Atom v2 l). split; eauto.
Qed.

Lemma increase_auth :
  forall pc e t1 t2 a1,
    pc; e ⊢ t1 ⇓ a1 ->
    term_auth_leq t1 t2 ->
    exists a2,
      pc; e ⊢ t2 ⇓ a2 /\ atom_auth_leq a1 a2.
Proof.
intros pc e t1 t2 a1 Heval Ht.
apply increase_auth_strong with (e1 := e) (t1 := t1); auto.
Qed.

Lemma increase_auth_bool :
  forall pc e t1 t2 b l,
    pc; e ⊢ t1 ⇓ Atom (VBool b) l ->
    term_auth_leq t1 t2 ->
    pc; e ⊢ t2 ⇓ Atom (VBool b) l.
Proof.
intros pc e t1 t2 b l Heval Ht.
pose proof (increase_auth _ _ _ t2 _ Heval Ht) as [[v' l'] [Heval' Hatom']].
destruct Hatom' as [Hv' Hl']. destruct v'; try tauto.
congruence.
Qed.

End Least.
