Require Import MyTactics.
Require Import MyList.
Require Import LabelAlgebra.
Require Import Lambda.
Require Import TermEquiv.
Require Import Auth.
Require Import LambdaNI.

Set Implicit Arguments.

(** * Lower bounds on atoms, values, environments. *)
Section LowerBound.

Context {A L: Type} {LA: LabelAlgebra A L}.
Context (auth: A) (lab: L).

Fixpoint atom_lower_bound (a: atom A L) : Prop :=
match a with
| Atom v l => value_lower_bound v /\ lab ⊑[auth] l
end
with value_lower_bound (v: value A L) : Prop :=
match v with
| VBool _ => True
| VClos e t => list_forall atom_lower_bound e
end.

Definition env_lower_bound := list_forall atom_lower_bound.

End LowerBound.

Hint Unfold atom_lower_bound value_lower_bound env_lower_bound.

(** * Compatibility with some other relations. *)
Section Compatibility.

Context {A L: Type} {LA: LabelAlgebra A L}.
Context (auth: A) (lab: L).

Lemma lower_bound_label_below :
  forall lab',
    lab' ⊑[auth] lab ->
    (forall a, atom_lower_bound auth lab  a ->
               atom_lower_bound auth lab' a)
    /\ (forall v, value_lower_bound auth lab  v ->
                  value_lower_bound auth lab' v)
    /\ (forall e, env_lower_bound auth lab  e ->
                  env_lower_bound auth lab' e).
Proof.
intros lab' Hlab. apply atom_value_env_mutind; intros.
* destruct H0. fold (value_lower_bound auth lab v) in H0. split.
  + fold (value_lower_bound auth lab' v). auto.
  + transitivity lab; trivial.
* compute. trivial.
* apply H. trivial.
* induction l; simpl in *; auto. split.
  + apply (H 0 a); tauto.
  + apply IHl; try tauto. intros n b Hn. apply (H (S n) b). assumption.
Qed.

Lemma atom_lower_bound_label_below :
  forall lab' a,
    lab' ⊑[auth] lab ->
    atom_lower_bound auth lab  a ->
    atom_lower_bound auth lab' a.
Proof.
intros lab' a H H0.
destruct (lower_bound_label_below lab' H) as [? _].
auto.
Qed.

Lemma value_lower_bound_label_below :
  forall lab' v,
    lab' ⊑[auth] lab ->
    value_lower_bound auth lab  v ->
    value_lower_bound auth lab' v.
Proof.
intros lab' v H H0.
destruct (lower_bound_label_below lab' H) as [_ [? _]].
auto.
Qed.

Lemma env_lower_bound_label_below :
  forall lab' e,
    lab' ⊑[auth] lab ->
    env_lower_bound auth lab  e ->
    env_lower_bound auth lab' e.
Proof.
intros lab' e H H0.
destruct (lower_bound_label_below lab' H) as [_ [_ ?]].
auto.
Qed.

Lemma atom_lower_bound_label_below0 :
  forall lab' a,
    lab' ⊑ lab ->
    atom_lower_bound auth lab  a ->
    atom_lower_bound auth lab' a.
Proof.
intros lab' a H H0.
apply atom_lower_bound_label_below; auto.
Qed.

Lemma value_lower_bound_label_below0 :
  forall lab' v,
    lab' ⊑ lab ->
    value_lower_bound auth lab  v ->
    value_lower_bound auth lab' v.
Proof.
intros lab' v H H0.
apply value_lower_bound_label_below; auto.
Qed.

Lemma env_lower_bound_label_below0 :
  forall lab' e,
    lab' ⊑ lab ->
    env_lower_bound auth lab  e ->
    env_lower_bound auth lab' e.
Proof.
intros lab' e H H0.
apply env_lower_bound_label_below; auto.
Qed.

Lemma lower_bound_equiv :
  (forall a a', atom_lower_bound auth lab a ->
                atom_equiv a a' ->
                atom_lower_bound auth lab a')
  /\ (forall v v', value_lower_bound auth lab v ->
                   value_equiv v v' ->
                   value_lower_bound auth lab v')
  /\ (forall e e', env_lower_bound auth lab e ->
                   env_equiv e e' ->
                   env_lower_bound auth lab e').
Proof.
apply atom_value_env_mutind; intros.
* destruct H0. fold (value_lower_bound auth lab v) in H0.
  destruct a' as [v' l']. destruct H1. fold (value_equiv v v') in H1.
  split.
  + fold (value_lower_bound auth lab v'). auto.
  + transitivity l; auto.
* destruct v'; compute in H0; tauto.
* destruct v'; try solve [compute in H1; tauto].
  destruct H1; apply H; assumption.
* generalize dependent e'.
  induction l; intros l' Henv; destruct l'; simpl in *; auto. split.
  + apply (H 0 a); tauto.
  + apply IHl; try tauto. intros n b Hn. apply (H (S n) b). assumption.
Qed.

Lemma atom_lower_bound_equiv :
  forall a a', atom_lower_bound auth lab a ->
               atom_equiv a a' ->
               atom_lower_bound auth lab a'.
Proof.
pose proof lower_bound_equiv. tauto.
Qed.

Lemma value_lower_bound_equiv :
  forall v v', value_lower_bound auth lab v ->
               value_equiv v v' ->
               value_lower_bound auth lab v'.
Proof.
pose proof lower_bound_equiv. tauto.
Qed.

Lemma env_lower_bound_equiv :
  forall e e', env_lower_bound auth lab e ->
               env_equiv e e' ->
               env_lower_bound auth lab e'.
Proof.
pose proof lower_bound_equiv. tauto.
Qed.

End Compatibility.

Hint Resolve
     atom_lower_bound_label_below
     value_lower_bound_label_below
     env_lower_bound_label_below
     atom_lower_bound_label_below0
     value_lower_bound_label_below0
     env_lower_bound_label_below0
     atom_lower_bound_equiv
     value_lower_bound_equiv
     env_lower_bound_equiv.

(** * Lower bound is preserved by evaluation. *)
Section LowerBoundPreservation.

Context {A L: Type}
        {LA: LabelAlgebra A L}
        (auth: A)
        (lab: L).

Lemma lower_bound_eval_lower_bound :
  forall pc e t a,
    pc; e ⊢ t ⇓ a ->
    lab ⊑[auth] pc ->
    env_lower_bound auth lab e ->
    env_auth e ≤ auth ->
    term_auth t ≤ auth ->
    atom_lower_bound auth lab a.
Proof.
intros pc e t a Heval HLBpc HLBenv HAenv HAt.
induction Heval; auto.
* pose proof (list_forall_atIndex _ _ _ _ HLBenv H) as [Hv Hl].
  split.
  + auto.
  + transitivity l; auto.
* simpl in HAt.
  assert (term_auth t1 ≤ auth) as HAt1 by eauto.
  assert (term_auth t2 ≤ auth) as HAt2 by eauto.
  clear HAt.
  specialize (IHHeval1 HLBpc HLBenv HAenv HAt1).
  destruct IHHeval1 as [He1' Hl1].
  fold (atom_lower_bound auth lab) in He1'.
  fold (env_lower_bound auth lab) in He1'.
  specialize (IHHeval2 HLBpc HLBenv HAenv HAt2).
  assert (lab ⊑[auth] l1) as HLBl1.
  { transitivity pc; trivial.
    eapply eval_pc_lower_bound with (t := t1); eauto.
  }
  assert (env_auth (a2 :: e1') ≤ auth) as HAa2e1'.
  { simpl.
    assert (env_auth e1' ≤ auth).
    { apply @eval_auth_upper_bound with (auth := auth) in Heval1; auto.
      unfold atom_auth in Heval1.
      fold atom_auth in Heval1. fold (env_auth e1') in Heval1.
      eauto.
    }
    assert (atom_auth a2 ≤ auth).
    { eapply @eval_auth_upper_bound with (e := e); eauto. }
    auto.
  }
  assert (term_auth t1' ≤ auth) as HAt1'.
  { apply @eval_auth_upper_bound with (auth := auth) in Heval1; auto.
    unfold atom_auth in Heval1.
    eauto.
  }
  specialize (IHHeval3 HLBl1 (conj IHHeval2 He1') HAa2e1' HAt1').
  assumption.
* simpl in HAt.
  destruct IHHeval as [Hv1 Hl1]; auto.
  + eauto.
  + fold (value_lower_bound auth lab) in Hv1. split; auto.
    transitivity l1; trivial. eauto.
Qed.

End LowerBoundPreservation.

(** * Sufficient condition for preservation of evaluation. *)
Section EvalPreservation.

Context {A1 L1 A2 L2: Type}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2)
        (lab: L1).

Lemma lower_bound_eval_preserved_strong :
  (forall auth l1 l2,
     lab ⊑[auth] l1 ->
     lab ⊑[auth] l2 ->
     Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2) ->
  (forall auth auth' l1 l2,
     lab ⊑[auth] l1 ->
     l1 ⊑[auth'] l2 ->
     Lmap m l1 ⊑[Amap m auth'] Lmap m l2) ->
  forall auth pc e t a,
    pc; e ⊢ t ⇓ a ->
    lab ⊑[auth] pc ->
    env_lower_bound auth lab e ->
    env_auth e ≤ auth ->
    term_auth t ≤ auth ->
    exists a',
      atom_equiv (apply m a) a'
      /\ Lmap m pc; apply m e ⊢ apply m t ⇓ a'.
Proof.
intros Hjoin HflowsTo auth pc e t a Heval HLBpc HLBenv HAenv HAt.
induction Heval.
* exists (apply m (Atom (VBool b) pc)).
  split.
  + reflexivity.
  + apply Eval_bool.
* exists (Atom (apply m v) (Lmap m l ⊔ Lmap m pc)).
  split.
  + split.
    - fold value_equiv. reflexivity.
    - apply (Hjoin auth); auto.
      apply (list_forall_atIndex _ _ _ _ HLBenv H).
  + apply Eval_var.
    replace (Atom (apply m v) (Lmap m l)) with (apply m (Atom v l))
      by reflexivity.
    apply atIndexMapSome; trivial.
* exists (apply m (Atom (VClos e t) pc)).
  split.
  + reflexivity.
  + apply Eval_lam.
* simpl in HAt.
  assert (term_auth t1 ≤ auth) as HAt1 by eauto.
  assert (term_auth t2 ≤ auth) as HAt2 by eauto.
  clear HAt.
  destruct IHHeval1 as [[v1' l1'] [Hatom1' Heval1']]; auto.
  destruct v1'; try solve [compute in Hatom1'; tauto].
  rename l into e1''. rename t into t1''.
  destruct Hatom1' as [[He1' Ht1'] Hl1'].
  fold atom_equiv in He1'. fold (LA_apply_atom m) in He1'.
  fold (LA_apply_env m) in He1'.
  replace (LA_apply_env m e1') with (apply m e1') in He1' by reflexivity.
  fold env_equiv in He1'.
  destruct IHHeval2 as [a2' [Hatom2' Heval2']]; auto.
  destruct IHHeval3 as [a3' [Hatom3' Heval3']].
  + transitivity pc; trivial. eapply eval_pc_lower_bound with (t := t1); eauto.
  + split.
    - apply @lower_bound_eval_lower_bound
      with (lab := lab) (auth := auth) in Heval2; auto.
    - apply @lower_bound_eval_lower_bound
      with (lab := lab) (auth := auth) in Heval1; auto.
      destruct Heval1; assumption.
  + simpl.
    apply @eval_auth_upper_bound with (auth := auth) in Heval1; auto.
    unfold atom_auth in Heval1.
    fold (@atom_auth _ _ LA1) in Heval1. fold (env_auth e1') in Heval1.
    apply @eval_auth_upper_bound with (auth := auth) in Heval2; auto.
    eauto.
  + apply @eval_auth_upper_bound with (auth := auth) in Heval1; auto.
    unfold atom_auth in Heval1.
    fold (@atom_auth _ _ LA1) in Heval1. fold (env_auth e1') in Heval1.
    eauto.
  + pose proof
         (@eval_equiv_compat_fun _ _ _
                                 (Lmap m l1) l1'
                                 (apply m (a2 :: e1')) (a2' :: e1'')
                                 (apply m t1') t1''
                                 a3') as H.
    destruct H as [a3'' [Heval3'' Hatom3'']]; auto.
    - split; auto.
    - { exists a3''. split.
        * transitivity a3'; auto.
        * apply @Eval_app with
          (e1' := e1'') (t1' := t1'') (l1 := l1') (a2 := a2'); auto.
      }
* simpl in HAt.
  destruct IHHeval as [[v1' l1'] [Hatom' Heval']]; auto. eauto.
  exists (Atom v1' (Lmap m l)).
  split.
  + destruct Hatom'. split; auto.
  + apply @Eval_relabel with (l1 := l1').
    - assumption.
    - destruct Hatom' as [_ Hl1'].
      transitivity (Lmap m l1); auto.
      apply (HflowsTo auth); auto.
      transitivity pc; auto.
      eapply eval_pc_lower_bound; eauto.
Qed.

Lemma lower_bound_eval_preserved :
  (forall auth l1 l2,
     lab ⊑[auth] l1 ->
     lab ⊑[auth] l2 ->
     Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2) ->
  (forall auth auth' l1 l2,
     lab ⊑[auth] l1 ->
     l1 ⊑[auth'] l2 ->
     Lmap m l1 ⊑[Amap m auth'] Lmap m l2) ->
  forall t a,
    lab; nil ⊢ t ⇓ a ->
    exists a',
      atom_equiv (apply m a) a'
      /\ Lmap m lab; nil ⊢ apply m t ⇓ a'.
Proof.
intros Hjoin HflowsTo t a Heval.
apply (lower_bound_eval_preserved_strong Hjoin HflowsTo (term_auth t) Heval);
  auto; simpl; auto.
Qed.

End EvalPreservation.

(** * [below] predicates. *)
Section Below.

Context {A L: Type} {LA: LabelAlgebra A L} (auth: A).

Fixpoint atom_below (a1 a2: atom A L) {struct a1} : Prop :=
match a1, a2 with
| Atom v1 l1, Atom v2 l2 =>
  value_below v1 v2 /\ l1 ⊑[auth] l2
end
with value_below (v1 v2: value A L) {struct v1} : Prop :=
match v1, v2 with
| VBool b1, VBool b2 => b1 = b2
| VClos e1 t1, VClos e2 t2 =>
  list_forall2 atom_below e1 e2 /\ term_equiv t1 t2
| VBool _, _ | VClos _ _, _ => False
end.

Definition env_below := list_forall2 atom_below.

End Below.

Hint Unfold env_below.

(** * [best] predicates *)
Section Best.

Context {A1 L1 A2 L2: Type}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2).

Fixpoint atom_best (a0: atom A2 L2) (a: atom A1 L1) : Prop :=
match a0, a with
| Atom v0 l0, Atom v l =>
  value_best v0 v
  /\ (forall auth' l',
        l0 ⊑[Amap m auth'] Lmap m l' ->
        l ⊑[auth'] l')
end
with value_best (v0: value A2 L2) (v: value A1 L1) : Prop :=
match v0, v with
| VBool _, VBool _ => True
| VClos e0 _, VClos e _ => list_forall2 atom_best e0 e
| VBool _, _ | VClos _ _, _ => False
end.

Definition env_best := list_forall2 atom_best.

End Best.

Hint Unfold env_best.

(** * [same_lab] predicates *)
Section Same.

Context {A L: Type} {LA: LabelAlgebra A L} (auth: A).

Fixpoint atom_same_lab (a1 a2: atom A L) {struct a1} : Prop :=
match a1, a2 with
| Atom v1 l1, Atom v2 l2 =>
  value_same_lab v1 v2 /\ l1 =L l2
end
with value_same_lab (v1 v2: value A L) {struct v1} : Prop :=
match v1, v2 with
| VBool _, VBool _ => True
| VClos e1 _, VClos e2 _ =>
  list_forall2 atom_same_lab e1 e2
| VBool _, _ | VClos _ _, _ => False
end.

Definition env_same_lab := list_forall2 atom_same_lab.

End Same.

Hint Unfold env_same_lab.

(** * Sufficient condition for reflection of evaluation. *)
Section EvalReflection.

Context {A1 L1 A2 L2: Type}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2)
        (lab: L2).

Lemma lower_bound_eval_reflected_strong :
  (forall auth auth' l1 l2,
     lab ⊑[Amap m auth] Lmap m l1 ->
     lab ⊑[Amap m auth] Lmap m l2 ->
     Lmap m l1 ⊔ Lmap m l2 ⊑[Amap m auth'] Lmap m (l1 ⊔ l2)) ->
  (forall auth auth' l1 l2,
     lab ⊑[Amap m auth] Lmap m l1 ->
     Lmap m l1 ⊑[Amap m auth'] Lmap m l2 ->
     l1 ⊑[auth'] l2) ->
  (forall auth auth' l1 l2,
     lab ⊑[Amap m auth] Lmap m l1 ->
     l1 ⊑[auth'] l2 ->
     Lmap m l1 ⊑[Amap m auth'] Lmap m l2) ->
  forall (lab': L1) (auth: A1) (pc: L2) (e: env A2 L2) (t: term A2 L2) a,
    (forall auth, lab ⊑[Amap m auth] Lmap m lab') ->
    pc; e ⊢ t ⇓ a ->
    forall (pc': L1) (e': env A1 L1) (t': term A1 L1),
    term_equiv t (apply m t') ->
    lab' ⊑[auth] pc' ->
    (forall auth', pc ⊑[Amap m auth'] Lmap m pc') ->
    (forall auth' pc'',
       pc ⊑[Amap m auth'] Lmap m pc'' ->
       pc' ⊑[auth'] pc'') ->
    env_lower_bound auth lab' e' ->
    (forall auth', env_below (Amap m auth') e (apply m e')) ->
    env_best m e e' ->
    env_auth e' ≤ auth ->
    term_auth t' ≤ auth ->
    exists a',
      pc'; e' ⊢ t' ⇓ a'
      /\ (forall auth', atom_below (Amap m auth') a (apply m a'))
      /\ atom_best m a a'.
Proof.
intros Hjoin HflowsTo HflowsTo2 lab' auth pc e t a Hlab Heval.
induction Heval;
  intros pc' e' t' Ht
         HLBpc Hpc Hbestpc
         HLBenv Henv Hbestenv
         HAenv HAt;
  destruct t'; try solve [compute in Ht; tauto].
* simpl in Ht. subst b0.
  exists (Atom (VBool b) pc'). splits; auto.
  + split; auto.
  + split; auto.
* simpl in Ht. subst n0.
  destruct (list_forall2_atIndex_fun _ _ _ _ _ (Henv auth) H)
    as [[v'' l''] [H'' Ha']].
  destruct Ha' as [Hv' Hl']. fold (value_below (Amap m auth)) in Hv'.
  pose proof (atIndexMapSome_inv _ _ _ _ H'') as [[v' l'] [H' Heq]].
  unfold LA_apply_atom in Heq. fold (LA_apply_value m) in Heq.
  inversion Heq; subst. clear Heq.
  replace (LA_apply_value m v') with (apply m v') in *|- by reflexivity.
  exists (Atom v' (l' ⊔ pc')). splits; auto.
  + intros auth'.
    destruct
      (list_forall2_atIndex_fun _ _ _ _ _ (Henv auth') H)
      as [[v0'' l0''] [H0'' [H0v'' H0l'']]].
    fold (value_below (Amap m auth')) in H0v''.
    assert (v0'' = apply m v') by congruence; subst v0''.
    assert (l0'' = Lmap m l') by congruence; subst l0''.
    clear H0''.
    split; auto.
    transitivity (Lmap m l' ⊔ Lmap m pc').
    - { repeat rewrite <- (joinBottom (Amap m auth')).
        apply join_lub.
        - transitivity (Lmap m l');  auto using join_ub1.
        - transitivity (Lmap m pc'); auto using join_ub2.
      }
    - { apply (Hjoin auth).
        - transitivity (Lmap m lab'); auto.
          pose proof (list_forall_atIndex _ _ _ _ HLBenv H') as [_ ?].
          apply (HflowsTo2 auth'); auto.
        - transitivity (Lmap m lab'); auto.
          apply (HflowsTo2 auth'); auto.
      }
  + split; auto.
    - fold (value_best m).
      pose proof (list_forall2_atIndex _ _ _ _ _ _ Hbestenv H H') as [? _].
      assumption.
    - { intros auth' l'' Hl''.
        rewrite <- (joinBottom auth').
        apply join_lub; auto.
        * pose proof (list_forall2_atIndex _ _ _ _ _ _ Hbestenv H H')
            as [_ Hlbest].
          apply Hlbest; auto.
          transitivity (l ⊔ pc); auto.
        * apply Hbestpc; auto. transitivity (l ⊔ pc); auto.
      }
* exists (Atom (VClos e' t') pc'). splits; auto.
  + intros auth'. split; auto.
    fold (atom_below (Amap m auth')).
    fold (LA_apply_atom m). fold (env_below (Amap m auth')).
    replace (map (LA_apply_atom m) e') with (apply m e') by reflexivity.
    split; eauto.
  + split; auto.
* simpl in HAt.
  assert (term_auth t'1 ≤ auth) as HAt1 by eauto.
  assert (term_auth t'2 ≤ auth) as HAt2 by eauto.
  clear HAt.
  destruct Ht as [Ht1 Ht2].
  destruct (IHHeval1 pc' e' t'1)
    as [[v1'' l1''] [Heval1' [Ha1'' Hbesta1'']]]; auto.
  clear IHHeval1.
  destruct v1'';
    try solve [specialize (Ha1'' auth); compute in Ha1''; tauto].
  rename l into e1''. rename t into t1''.
  destruct (IHHeval2 pc' e' t'2)
    as [a2' [Heval2' [Ha2' Hbesta2']]]; auto. clear IHHeval2.
  destruct (IHHeval3 l1'' (a2' :: e1'') t1'')
    as [a3' [Heval3' [Ha3' Hbesta3']]]; clear IHHeval3.
  + specialize (Ha1'' auth). destruct Ha1''. tauto.
  + transitivity pc'; auto.
    apply (eval_pc_lower_bound _ HAenv HAt1 Heval1').
  + intros auth'. specialize (Ha1'' auth').
    destruct Ha1''; auto.
  + destruct Hbesta1''; assumption.
  + split; auto.
    - apply @lower_bound_eval_lower_bound with (pc := pc') (e := e') (t := t'2);
      auto.
    - pose proof (lower_bound_eval_lower_bound
                    _ _ Heval1' HLBpc HLBenv HAenv HAt1) as [? _].
      assumption.
  + intros auth'.
    split; auto.
    specialize (Ha1'' auth'). destruct Ha1'' as [[? _] _]. assumption.
  + split; auto. destruct Hbesta1'' as [? _]. assumption.
  + simpl. apply join_lub.
    - pose proof (eval_auth_upper_bound _ HAenv HAt1 Heval1') as HAa1'.
      unfold atom_auth in HAa1'.
      fold (@atom_auth _ _ LA1) in HAa1'. fold (env_auth e1'') in HAa1'.
      eauto.
    - apply (eval_auth_upper_bound _ HAenv HAt2 Heval2').
  + pose proof (eval_auth_upper_bound _ HAenv HAt1 Heval1') as HAa1'.
    unfold atom_auth in HAa1'.
    fold (@atom_auth _ _ LA1) in HAa1'. fold (env_auth e1'') in HAa1'.
    eauto.
  + exists a3'. splits; auto.
    apply @Eval_app with (e1' := e1'') (t1' := t1'') (l1 := l1'') (a2 := a2');
    auto.
* rename a0 into a'. rename l0 into l'.
  destruct Ht as [Ht [Ha Hl]].
  simpl in HAt.
  assert (term_auth t' ≤ auth) as HAt1 by eauto.
  assert (a' ≤ auth) as Ha2 by eauto.
  clear HAt. rename HAt1 into HAt.
  specialize (IHHeval _ _ _ Ht HLBpc Hpc Hbestpc HLBenv Henv Hbestenv HAenv HAt).
  destruct IHHeval as [[v1' l1'] [Heval' [Ha1' Hbesta1']]].
  exists (Atom v1' l'). splits; auto.
  + apply @Eval_relabel with (l1 := l1'); auto.
    destruct Hbesta1' as [_ Hbestl1'].
    apply Hbestl1'; auto.
    transitivity l; auto. eauto.
  + intros auth'.
    split; auto.
    fold (value_below (Amap m auth')).
    fold (LA_apply_value m).
    replace (LA_apply_value m v1') with (apply m v1') by reflexivity.
    destruct (Ha1' auth'); auto.
  + split.
    - fold (value_best m). destruct Hbesta1'; auto.
    - { destruct Hbesta1' as [_ Hbestl1'].
        intros auth' l1'' Hl1''.
        apply (HflowsTo auth).
        * transitivity (Lmap m lab'); auto.
          apply (HflowsTo2 a'); auto.
          transitivity pc'; auto. transitivity l1'.
          apply (eval_pc_lower_bound _ HAenv HAt Heval').
          apply (flowsToIncluded a' auth); auto.
          apply Hbestl1'; auto. transitivity l; auto. eauto.
        * transitivity l; auto.
      }
Qed.

Lemma lower_bound_eval_reflected :
  (forall (auth0 auth' : A1) (l1 l2 : L1),
     lab ⊑[ Amap m auth0]Lmap m l1 ->
     lab ⊑[ Amap m auth0]Lmap m l2 ->
     Lmap m l1 ⊔ Lmap m l2 ⊑[ Amap m auth']Lmap m (l1 ⊔ l2)) ->
  (forall auth auth' l1 l2,
     lab ⊑[Amap m auth] Lmap m l1 ->
     Lmap m l1 ⊑[Amap m auth'] Lmap m l2 ->
     l1 ⊑[auth'] l2) ->
  (forall auth auth' l1 l2,
     lab ⊑[Amap m auth] Lmap m l1 ->
     l1 ⊑[auth'] l2 ->
     Lmap m l1 ⊑[Amap m auth'] Lmap m l2) ->
  forall (pc': L1) (t': term A1 L1) a,
    lab; nil ⊢ (apply m t') ⇓ a ->
    (forall auth, lab ⊑[Amap m auth] Lmap m pc') ->
    (forall auth pc'',
       lab ⊑[Amap m auth] Lmap m pc'' ->
       pc' ⊑[auth] pc'') ->
    exists a',
      pc'; nil ⊢ t' ⇓ a'
      /\ (forall auth',
            atom_below (Amap m auth') a (apply m a'))
      /\ atom_best m a a'.
Proof.
intros Hjoin HflowsTo HflowsTo2 pc' t' a Heval HLBpc Hbestpc.
apply lower_bound_eval_reflected_strong
with (lab' := pc') (pc := lab) (e := nil) (t := apply m t') (auth := term_auth t');
  simpl; auto.
Qed.

End EvalReflection.
