Require Import MyClassical.
Require Import Ensembles.
Require Import Constructive_sets.
Require Import Classical_sets.
Require Import Finite_sets.
Require Finite_sets_facts.
Require Image.
Require Import SetoidClass.

(** * Some facts about sets. *)

(** ** Union with empty set. *)
Lemma Union_with_empty_set_Included1 {A: Type} (s1 s2: Ensemble A) :
  Included _ s1 (Union _ s2 (Empty_set _)) <-> Included _ s1 s2.
Proof.
split; intros H x Hx.
* specialize (H x Hx). apply Union_inv in H. destruct H as [H | []]; trivial.
* left. apply H. trivial.
Qed.

Lemma Union_with_empty_set_Included2 {A: Type} (s1 s2: Ensemble A) :
  Included _ (Union _ s1 (Empty_set _)) s2 <-> Included _ s1 s2.
Proof.
split; intros H x Hx.
* assert (In _ (Union _ s1 (Empty_set _)) x) as Hx'. { left. trivial. }
  specialize (H x Hx'). assumption.
* apply Union_inv in Hx. destruct Hx as [Hx | []]. apply H. trivial.
Qed.

(** ** Distributivity *)
Lemma Ensemble_distrib1 {A: Type} (s1 s2 s3: Ensemble A) :
  Included _
    (Intersection _ (Union _ s1 s3) (Union _ s2 s3))
    (Union _ (Intersection _ s1 s2) s3).
Proof.
intros x H.
apply Intersection_inv in H. destruct H as [H13 H23].
apply Union_inv in H13.
apply Union_inv in H23.
destruct H13 as [H1 | H3].
* destruct H23 as [H2 | H3].
  + left. constructor; assumption.
  + right. assumption.
* right. assumption.
Qed.

Lemma Ensemble_distrib2 {A: Type} (s1 s2 s3: Ensemble A) :
  Included _
    (Union _ (Intersection _ s1 s2) s3)
    (Intersection _ (Union _ s1 s3) (Union _ s2 s3)).
Proof.
intros x H.
apply Union_inv in H. destruct H as [H12 | H3].
* apply Intersection_inv in H12. destruct H12 as [H1 H2].
  constructor; left; assumption.
* constructor; right; assumption.
Qed.

(** ** Complementation *)
Lemma Complement_Included {A} (s1 s2: Ensemble A) :
  Included _ s1 s2 <-> Included _ (Complement _ s2) (Complement _ s1).
Proof.
split; intros H a.
* intros Ha Hn. apply Ha. clear Ha. auto.
* apply contraposition_not. intro Ha. apply H. assumption.
Qed.

Lemma Complement_injective {A} : Image.injective _ _ (Complement A).
Proof.
intros s1 s2 H.
apply Extensionality_Ensembles.
split; intros x; apply contraposition_not; intro Hx.
* assert (In A (Complement A s2) x) as H0 by assumption.
  rewrite <- H in H0. assumption.
* assert (In A (Complement A s1) x) as H0 by assumption.
  rewrite H in H0. assumption.
Qed.

Lemma Complement_Empty_set {A: Type} :
  Complement _ (Empty_set A) = Full_set _.
Proof.
apply Extensionality_Ensembles. split; intros a H.
* apply Full_intro.
* intros [].
Qed.

Lemma Complement_Full_set {A: Type} :
  Complement _ (Full_set A) = Empty_set _.
Proof.
apply Extensionality_Ensembles. split; intros a H.
* exfalso. apply H. apply Full_intro.
* destruct H.
Qed.

(** ** De Morgan laws. *)
Lemma Complement_Union {A: Type} (S1 S2: Ensemble A):
  Complement _ (Union _ S1 S2)
  = Intersection _ (Complement _ S1) (Complement _ S2).
Proof.
apply Extensionality_Ensembles. split; intros a H.
* split; intro Ha; apply H; [left | right]; assumption.
* apply Intersection_inv in H. destruct H as [H1 H2].
  intro Ha. apply Union_inv in Ha.
  destruct Ha; [apply H1 | apply H2]; assumption.
Qed.

Lemma Complement_Intersection {A: Type} (S1 S2: Ensemble A):
  Complement _ (Intersection _ S1 S2)
  = Union _ (Complement _ S1) (Complement _ S2).
Proof.
apply Extensionality_Ensembles. split; intros a H.
* apply NNPP. intro Ha. apply H. clear H.
  assert (In _ (Complement _ (Union _ (Complement _ S1) (Complement _ S2))) a)
    as Ha' by assumption. clear Ha.
  rewrite Complement_Union in Ha'.
  apply Intersection_inv in Ha'.
  repeat rewrite Complement_Complement in Ha'.
  split; tauto.
* intro Ha. apply Intersection_inv in Ha. apply Union_inv in H.
  destruct H as [H | H]; apply H; tauto.
Qed.

(** ** [Complement] as an adjoint. *)
Lemma Complement_adjoint {A: Type} (S1 S2 S3: Ensemble A) :
  Included _ S1 (Union _ S2 S3)
  <-> Included _ (Intersection _ S1 (Complement _ S3)) S2.
Proof.
split; intros H a Ha.
* apply Intersection_inv in Ha. destruct Ha as [Ha1 Ha3].
  specialize (H _ Ha1). apply Union_inv in H.
  destruct H; [assumption | exfalso; apply Ha3; assumption].
* destruct (classic (In A S3 a)).
  + right. assumption.
  + left. apply (H a). split; assumption.
Qed.

Lemma Complement_adjoint' {A: Type} (S1 S2 S3: Ensemble A) :
  Included _ S1 (Union _ S2 (Complement _ S3))
  <-> Included _ (Intersection _ S1 S3) S2.
Proof.
rewrite Complement_adjoint.
rewrite Complement_Complement.
reflexivity.
Qed.

(** ** Subsets *)
Lemma subsets_of_Empty_set (A: Type) :
  Same_set _
           (fun S : Ensemble A => Included _ S (Empty_set A))
           (Singleton _ (Empty_set A)).
Proof.
split.
* intros S H. auto with sets.
* intros S H. unfold In.
  apply Singleton_inv in H. subst. auto with sets.
Qed.

Lemma subsets_of_Add (A: Type) (a : A) (S : Ensemble A) :
   Same_set _
            (fun S' : Ensemble A => Included A S' (Add A S a))
            (Union _
                   (fun S' : Ensemble A => Included A S' S)
                   (Image.Im _ _ (fun S' : Ensemble A => Included A S' S)
                       (fun S => Add A S a))).
Proof.
split; intros S' HS'.
* unfold In in HS'.
  destruct (classic (In A S' a)) as [Ha | Ha].
  + right.
    replace S' with (Add _ (Subtract _ S' a) a) by auto with sets.
    apply Image.Im_def with (x := Subtract A S' a).
    unfold In. intros x Hx.
    apply Subtract_inv in Hx. destruct Hx as [Hx Hax].
    specialize (HS' x Hx).
    apply Union_inv in HS'. destruct HS' as [HS' | HS']; auto.
    apply Singleton_inv in HS'. congruence.
  + left. unfold In. intros x Hx.
    specialize (HS' x Hx).
    apply Union_inv in HS'. destruct HS' as [HS' | HS']; auto.
    apply Singleton_inv in HS'. congruence.
* unfold In. apply Union_inv in HS'. destruct HS' as [HS' | HS'].
  + unfold In in HS'. auto with sets.
  + apply Image.Im_inv in HS'. destruct HS' as [S'' [HS'' Heq]].
    subst S'. unfold In in HS''. auto with sets.
Qed.

Lemma Finite_subsets (A: Type) :
  forall S : Ensemble A,
    Finite A S ->
    Finite (Ensemble A) (fun S' => Included _ S' S).
Proof.
intros S HS. induction HS.
* apply (Finite_sets_facts.Finite_downward_closed _ (Singleton _ (Empty_set _))).
  + apply (Finite_sets_facts.Finite_downward_closed _
             (Add _ (Empty_set _) (Empty_set _))).
    auto with sets.
    intros S H. apply Singleton_inv in H. subst. auto with sets.
  + destruct (subsets_of_Empty_set A). assumption.
* rename A0 into S. rename x into a.
  apply (Finite_sets_facts.Finite_downward_closed _
           (Union _
                  (fun S' : Ensemble A => Included A S' S)
                  (Image.Im _ _ (fun S' : Ensemble A => Included A S' S)
                      (fun S => Add A S a)))).
  + apply Finite_sets_facts.Union_preserves_Finite; auto using Image.finite_image.
  + destruct (subsets_of_Add _ a S). assumption.
Qed.
