Require Import MyTactics.
Require Import MyList.

(** * Preliminary definitions *)
Definition list_apply {A} (Ps: list (A -> Prop)) (l: list A) : Prop :=
  list_forall2 (fun P x => P x) Ps l.

Definition table_apply {A}
           (Pt: list (list (A -> Prop)))
           (t: list (list A)) : Prop :=
  list_forall2 list_apply Pt t.

Inductive Answer : Type := Yes | No | Unknown.
Notation "✓" := Yes : table.
Notation "×" := No : table.
Notation "?" := Unknown : table.
Notation "[ x ; .. ; y ]" := (cons x .. (cons y nil) ..) : table.

Definition lift_binary_predicate {A B : Type} (P: A -> B -> Prop) :
  A -> B -> Answer -> Prop :=
  fun a b ans =>
    match ans with
      | Yes => P a b
      | No => not (P a b)
      | Unknown => True
    end.

Definition make_prop_table {A B} (P: A -> B -> Prop)
           (rows: list A) (cols: list B) : list (list (Answer -> Prop)) :=
  map (fun r => map (fun c => lift_binary_predicate P r c) cols) rows.

(** * Main functions *)
(** This is the function to use to check non square tables. *)
Definition property_table {A B: Type}
           (P: A -> B -> Prop)
           (rows: list A) (cols: list B)
           (answers: list (list Answer)) : Prop :=
  table_apply (make_prop_table P rows cols) answers.

Lemma property_table_conjs {A B: Type}
      (P: A -> B -> Prop)
      (rows: list A) (cols: list B)
      (answers: list (list Answer)) :
  conjs
    (flat_map2 (map2 (fun P x => P x) False)
               False
               (make_prop_table P rows cols)
               answers) ->
  property_table P rows cols answers.
Proof.
apply conjs_list_forall2_bis.
Qed.

(** This is the function to use to check square tables. *)
Definition property_square_table {A: Type}
           (P: A -> A -> Prop)
           (elts: list A)
           (answers: list (list Answer)) : Prop :=
  property_table P elts elts answers.

(** [split_property_table] unfolds what it means for a tabled property to hold and splits the goal into subgoals. *)
Ltac split_property_table :=
  apply property_table_conjs; simpl; splits.

(** * Correctness of the table predicates. *)
Lemma property_table_correct {A B: Type}
           (P: A -> B -> Prop)
           (rows: list A) (cols: list B)
           (answers: list (list Answer)) :
  property_table P rows cols answers ->
  forall n m a b,
    atIndex rows n = Some a ->
    atIndex cols m = Some b ->
    exists row answer,
      atIndex answers n = Some row /\
      atIndex row m = Some answer /\
      (answer = Yes -> P a b) /\
      (answer = No -> ~ P a b).
Proof.
revert cols answers.
induction rows; intros cols answers Htable n m r c Hn Hm;
unfold property_table, table_apply, make_prop_table in Htable; auto.
destruct answers as [| row answers].
simpl in Htable; tauto.
destruct n; simpl in Hn.
* inversion Hn; subst a; clear Hn. simpl in Htable.
  destruct Htable as [Hrow Htable].
  unfold list_apply in Hrow.
  pose proof
       (list_forall2_atIndex_fun
          _ (map (lift_binary_predicate P r) cols)
          row m
          (lift_binary_predicate P r c)
          Hrow) as [answer [Hrowm Hanswer]].
  apply atIndexMapSome; auto.
  exists row, answer.
  splits; solve [auto |
                 intro Heq; rewrite Heq in Hanswer;
                 unfold lift_binary_predicate in Hanswer; assumption].
* simpl in Htable.
  destruct Htable as [Hrow Htable].
  specialize (IHrows cols answers Htable n m r c Hn Hm).
  destruct IHrows as [row' [answer' [Hn' [Hm' [HYes HNo]]]]].
  exists row', answer'. simpl. auto.
Qed.

Lemma property_table_square_correct {A: Type}
           (P: A -> A -> Prop)
           (elts: list A)
           (answers: list (list Answer)) :
  property_square_table P elts answers ->
  forall n m a b,
    atIndex elts n = Some a ->
    atIndex elts m = Some b ->
    exists row answer,
      atIndex answers n = Some row /\
      atIndex row m = Some answer /\
      (answer = Yes -> P a b) /\
      (answer = No -> ~ P a b).
Proof.
apply property_table_correct.
Qed.

(** * Examples *)
Module Examples.

Open Scope table.

Goal property_table
     lt
     [ 1; 2 ]
     [ 0; 1; 2 ]
     [ [ ×; ×; ✓ ]
     ; [ ×; ×; × ]
     ].
Proof.
split_property_table; auto.
Qed.

Goal property_square_table
     lt
     [ 0; 1; 2 ]
     [ [ ×; ✓; ✓ ]
     ; [ ×; ×; ✓ ]
     ; [ ×; ×; × ]
     ].
Proof.
split_property_table; auto.
Qed.

Goal property_square_table
     lt
     [ 0; 1; 2 ]
     [ [ ✓; ✓; ✓ ]
     ; [ ×; ×; ✓ ]
     ; [ ×; ×; × ]
     ]
     -> False.
Proof.
compute. auto.
Qed.

Goal property_square_table
     lt
     [ 0; 1; 2 ]
     [ [ ×; ✓; ✓ ]
     ; [ ×; ×; ✓ ]
     ; [ ×; ×]
     ]
     -> False.
Proof.
compute. auto.
Qed.

Goal property_square_table
     lt
     [ 0; 1; 2 ]
     [ [ ×; ✓; ✓ ]
     ; [ ×; ×; ✓ ]
     ]
     -> False.
Proof.
compute. auto.
Qed.

Goal property_square_table
     lt
     [ 0; 1; 2 ]
     nil
     -> False.
Proof.
compute. auto.
Qed.

Goal property_square_table
     lt
     [ 0; 1; 2 ]
     [ nil ]
     -> False.
Proof.
compute. auto.
Qed.

Close Scope table.

End Examples.

(** * Exporting tables to other formats (TeX, HTML,...). *)
Module ExportTo.

Require Import String.

(** That axiom is used only to provide a fake printing function to Coq. It is realized in OCaml by a printing function. *)
Axiom print_string : string -> unit.

Open Scope string.
Open Scope table.

Definition list_to_string {A} (open del close: string)
           (f: A -> string) (l: list A) : string :=
  let loop :=
      fix loop (l: list A) : string :=
        match l with
          | nil => ""
          | a :: nil => f a
          | a :: l => f a ++ del ++ loop l
        end
  in
  open ++ loop l ++ close.

Goal list_to_string "[" "; " "]" (fun s => s) ["a"; "b";"c"] = "[a; b; c]".
Proof.
reflexivity.
Qed.

Goal list_to_string "[" "; " "]" (fun s => s) ["a"] = "[a]".
Proof.
reflexivity.
Qed.

Goal list_to_string "[" "; " "]" (fun s => s) nil = "[]".
Proof.
reflexivity.
Qed.

Definition table_to_string {A B C}
           (ropen rdel rclose: string) (rf: A -> string)
           (rows: list A)
           (copen cdel cclose: string) (cf: B -> string)
           (cols: list B)
           (topen tdel tclose: string) (tf: C -> string)
           (table: list (list C)) : string :=
  list_to_string
    (topen ++ list_to_string copen cdel cclose cf cols)
    tdel
    tclose
    (fun r =>
         list_to_string
           (ropen ++ rf (fst r))
           rdel
           rclose
           tf
           (snd r))
    (combine pair rows table).

Module HTML.

Definition export_answer ans :=
match ans with
| Yes => "✓"
| No => "✗"
| Unknown => "?"
end.

Definition export {A B}
           (rf: A -> string) (cf: B -> string)
           (rows: list A) (cols: list B)
           (table: list (list Answer)) : string :=
  table_to_string
     "<tr>" "" "</tr>" (fun r => "<th>" ++ rf r ++ "</th>") rows
     "<thead><tr><td></td>" "" "</tr></thead><tbody>" (fun c => "<th>" ++ cf c ++ "</th>") cols
     "<table>" "" "</tbody></table>" (fun a => "<td>" ++ export_answer a ++ "</td>")
     table.

Definition export_simple := export (fun s: string => s) (fun s: string => s).

Definition print {A B}
           (rf: A -> string) (cf: B -> string)
           (rows: list A) (cols: list B)
           (table: list (list Answer)) : unit :=
  print_string (export rf cf rows cols table).

Definition print_simple := print (fun s: string => s) (fun s: string => s).

End HTML.

Module TeX.

(** The TeX document should contain the command "\includepackage{pifont}". *)

Open Scope char_scope.
Definition newline := String "010" EmptyString.
Close Scope char_scope.

Definition export_answer ans :=
match ans with
| Yes => "\ding{51}"
| No => "\ding{55}"
| Unknown => "?"
end.

Definition export {A B}
           (rf: A -> string) (cf: B -> string)
           (rows: list A) (cols: list B)
           (table: list (list Answer)) : string :=
  let format := list_to_string "{c|" "" ("}" ++ newline) (fun _ => "c") cols
  in
  table_to_string
     "" " & " (" \\" ++ newline) (fun r => rf r ++ " & ") rows
     " & " " & " (" \\ \hline" ++ newline) cf cols
     ("\begin{tabular}" ++ format) "" "\end{tabular}" export_answer
     table.

Definition export_simple := export (fun s: string => s) (fun s: string => s).

Definition print {A B}
           (rf: A -> string) (cf: B -> string)
           (rows: list A) (cols: list B)
           (table: list (list Answer)) : unit :=
  print_string (export rf cf rows cols table).

Definition print_simple := print (fun s: string => s) (fun s: string => s).

End TeX.

Close Scope string.

End ExportTo.

Require Import Ascii String.
Extraction Language Ocaml.
Extract Inductive unit => "unit" [ "()" ].
Extract Inductive bool => "bool" [ "true" "false" ].
Extract Inductive list => "list" [ "[]" "(::)" ].
Extract Inductive prod => "(*)"  [ "(,)" ].
Extract Inductive ascii => char
[
"(* If this appears, you're using Ascii internals. Please don't *) (fun (b0,b1,b2,b3,b4,b5,b6,b7) -> let f b i = if b then 1 lsl i else 0 in Char.chr (f b0 0 + f b1 1 + f b2 2 + f b3 3 + f b4 4 + f b5 5 + f b6 6 + f b7 7))"
]
"(* If this appears, you're using Ascii internals. Please don't *) (fun f c -> let n = Char.code c in let h i = (n land (1 lsl i)) <> 0 in f (h 0) (h 1) (h 2) (h 3) (h 4) (h 5) (h 6) (h 7))".
Extract Constant ExportTo.print_string =>
"let rec print_string = function EmptyString -> print_newline () | String (c, s) -> begin print_char c; print_string s end in print_string".

(*
Open Scope table. Open Scope string.
Definition print_myTable :=
  ExportTo.TeX.print_simple
    [ "1"; "2" ]
    [ "3"; "4"; "5" ]
    [ [ ×; ×; ✓ ]
    ; [ ×; ×; × ]
    ].
Extraction "test.ml" print_myTable.
*)
